
# 分析用のキューブ

class Cube
  # new 次元
  def initialize dim
    raise ArgumentError if dim < 1
    @dim = dim
    @axes = []
    @data = []
  end
  
  # add [m, n, ...], val
  # valは整数でなければならない
  def add keys, val
    raise ArgumentError if keys.size != @dim
    raise TypeError if !val.is_a?(Integer)
    
    ary = @data
    keys.each_with_index {|key, i|
      axis = axis_map i
      axis[key] = axis.size if !axis[key]
      if i != @dim - 1
        ary[axis[key]] = [] if !ary[axis[key]]
        ary = ary[axis[key]]
      else
        ary[axis[key]] = ary[axis[key]].to_i + val
      end
    }
  end
  
  # 値、あるいは値の合計
  def [](*keys)
    raise ArgumentError if keys.size != @dim
    return sum_or_nil(0, @data, keys)
  end
  
  # 軸のキーの配列
  def axis n
    @axes[n].keys.sort {|x, y| x <=> y}
  end

  private
  def sum_or_nil i, ary_or_val, keys
    raise "i = #{i}, keys = #{keys.inspect}, ary = #{ary_or_val.inspect}" if keys.size + i != @dim
    return ary_or_val if keys.empty?

    axis = axis_map i
    keys = keys.dup
    key = keys.shift
    if !key
      val = 0
      ary_or_val.each {|x|
        if (r = sum_or_nil(i + 1, x, keys))
          val += r
        end
      }
      return val
    else
      return sum_or_nil(i + 1, ary_or_val[axis[key]], keys)
    end
  end
  
  # 軸ごとのkey => index のハッシュ
  # n    軸。0..(dim - 1)
  def axis_map n
    raise ArgumentError if n < 0 || n >= @dim
    @axes[n] = {} if !@axes[n]
    return @axes[n] 
  end
end


if __FILE__ == $0
  c = Cube.new 3
  c.add [:x, 1, 'a'], 10
  c.add [:x, 1, 'hoge'], 20
  c.add [:z, 1, 'a'], 30
  p c[:x, 1, 'hoge']  #=> 20
  p c[:z, 1, 'hoge']  #=> nil
  p c[nil, nil, 'a']  #=> 40
end
